library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.pcie.all;
use work.chipscope_h.all;
use work.wb.all;


-- interface to request application
entity exzess_pcie_app_if is
	port (
		mem_rd_i: in mem_rd_o_type; -- for the receive direction, data type names are _reversed_
		mem_rd_rbar_hit_i: in std_logic_vector(6 downto 0);
		mem_rd_o: out mem_rd_i_type;

		mem_wr_i: in mem_wr_o_type;
		mem_wr_rbar_hit_i: in std_logic_vector(6 downto 0);
		mem_wr_o: out mem_wr_i_type;

		cpl_i: in cpl_o_type;
		cpl_o: out cpl_i_type;
		cpld_i: in cpld_o_type;
		cpld_o: out cpld_i_type;

		tx_mem_rd_o: out mem_rd_o_type;
		tx_mem_rd_i: in mem_rd_i_type;

		tx_mem_wr_o: out mem_wr_o_type;
		tx_mem_wr_i: in mem_wr_i_type;

		tx_cpld_o: out cpld_o_type;
		tx_cpld_i: in cpld_i_type;

		app_request_o: out wb_master_o;
		app_request_i: in wb_master_i;
		
		mem_request_i: in wb_slave_i;
		mem_request_o: out wb_slave_o;

		base_address: out std_logic_vector(31 downto 0);

		clk: in std_logic;
		reset_n: in std_logic--;
	);
end entity exzess_pcie_app_if;

architecture exzess_pcie_app_if_arch of exzess_pcie_app_if is
	signal cs_ctrl: std_logic_vector(35 downto 0);
	signal cs_trig0: std_logic_vector(127 downto 0);

	signal dma_base_address_reg: std_logic_vector(31 downto 0);
	signal dma_region_size: std_logic_vector(31 downto 0);
	signal xor_pattern: std_logic_vector(31 downto 0);

	signal incoming_read_count: std_logic_vector(31 downto 0);
	signal incoming_write_count: std_logic_vector(31 downto 0);
	signal incoming_cpld_count: std_logic_vector(31 downto 0);
	signal outgoing_read_count: std_logic_vector(31 downto 0);
	signal outgoing_write_count: std_logic_vector(31 downto 0);
	signal outgoing_cpld_count: std_logic_vector(31 downto 0);

	signal in_rst_state: std_logic;
	signal in_out_read_state: std_logic;
	signal in_out_write_state: std_logic;
	signal in_in_cpld_state: std_logic;
	signal in_out_cpld_state: std_logic;
	signal in_wait_cpld_state: std_logic;
	signal in_wait_a_bit_state: std_logic;
	signal in_wait_for_ack_state: std_logic;

	signal app_request_o_reg: wb_master_o;
	signal mem_request_o_reg: wb_slave_o;

	type transformation_state_type is (rst,
		outgoing_read, outgoing_write,
		incoming_cpld, outgoing_cpld,
		wait_cpld, wait_a_bit, wait_for_ack
	);
	signal transformation_state: transformation_state_type;
	signal general_busy: std_logic;

	signal mem_rd_o_ack_buf: std_logic;
	signal mem_wr_o_ack_buf: std_logic;
	signal cpld_o_ack_buf: std_logic;
	signal cpl_o_ack_buf: std_logic;

	signal relevant_adr: std_logic_vector(31 downto 0);
	signal relevant_dat: std_logic_vector(31 downto 0);

	function incr(x: in std_logic_vector) return std_logic_vector is
	begin
		return std_logic_vector(unsigned(x) + 1);
	end function incr;

	--! swap all bytes in a DW, so 4321 will become 1234.
	--! Useful to convert little to big endian and vice versa.
	function bswap32(x: in std_logic_vector(31 downto 0)) return std_logic_vector is
	begin
		return x(7 downto 0) & x(15 downto 8) & x(23 downto 16) & x(31 downto 24);
	end function bswap32;
begin

	cs_icon: chipscope_icon port map (control0 => cs_ctrl);
	cs_ila: chipscope_ila port map (control => cs_ctrl, clk => clk, trig0 => cs_trig0);

	cs_trig0(1 downto 0) <= "10";
	cs_trig0(2) <= reset_n;
	cs_trig0(3) <= mem_rd_i.valid;
	cs_trig0(4) <= in_rst_state;
	cs_trig0(5) <= in_out_read_state;
	cs_trig0(6) <= in_out_write_state;
	cs_trig0(7) <= in_in_cpld_state;
	cs_trig0(8) <= in_out_cpld_state;
	cs_trig0(9) <= in_wait_cpld_state;
	cs_trig0(17) <= in_wait_a_bit_state;
	cs_trig0(18) <= in_wait_for_ack_state;
	cs_trig0(19) <= mem_request_i.we;
	cs_trig0(16 downto 10) <= mem_rd_rbar_hit_i;
	cs_trig0(49 downto 20) <= mem_rd_i.tlp.addr;
	cs_trig0(50) <= mem_request_i.cyc;
	cs_trig0(51) <= mem_request_i.stb;
	cs_trig0(53) <= mem_request_o_reg.ack;

	cs_trig0(52) <= app_request_i.ack;
	cs_trig0(54) <= app_request_o_reg.cyc;
	cs_trig0(55) <= app_request_o_reg.stb;

	cs_trig0(91 downto 60) <= relevant_adr;
	cs_trig0(123 downto 92) <= relevant_dat;

	cs_trig0(124) <= mem_rd_o_ack_buf;
	cs_trig0(125) <= mem_wr_o_ack_buf;
	cs_trig0(126) <= cpld_o_ack_buf;
	cs_trig0(127) <= cpl_o_ack_buf;

	base_address <= dma_base_address_reg;
	mem_request_o <= mem_request_o_reg;
	app_request_o <= app_request_o_reg;

	mem_wr_o.busy <= general_busy;
	mem_rd_o.busy <= general_busy;
	cpl_o.busy <= '0';

	mem_rd_o.ack <= mem_rd_o_ack_buf; -- buffered, so we can read from it
	mem_wr_o.ack <= mem_wr_o_ack_buf;
	cpld_o.ack <= cpld_o_ack_buf;
	cpl_o.ack <= cpl_o_ack_buf;

	handle_config_registers: process(clk, reset_n, mem_rd_i) is
	begin
		if (clk = '1' and clk'event) then
			if (reset_n = '0') then
			else
			end if;
		end if;
	end process handle_config_registers;

	handle_data_transformation: process(clk, reset_n, mem_rd_i, mem_wr_i) is
		variable mem_tlp: mem_tlp;

		variable write_enable: std_logic := '0';
		variable data: std_logic_vector(31 downto 0) := (others => '0');
		variable address: std_logic_vector(31 downto 0) := (others => '0');
		variable transaction_counter: std_logic_vector(5 downto 0);

		variable original_requester_id: std_logic_vector(15 downto 0);
		variable original_tag: std_logic_vector(7 downto 0);
		variable original_lower_addr: std_logic_vector(4 downto 0);
		variable mem_ack_sent: std_logic;
		variable app_ack_received: std_logic;
	begin
		if (clk = '1' and clk'event) then
			if (reset_n = '0') then

				in_rst_state <= '0';
				in_out_read_state <= '0';
				in_out_write_state <= '0';
				in_in_cpld_state <= '0';
				in_out_cpld_state <= '0';
				in_wait_cpld_state <= '0';
				in_wait_a_bit_state <= '0';

				mem_tlp := null_mem_tlp;

				-- mem_rd_o <= null_mem_rd_i_type;
				-- mem_wr_o <= null_mem_wr_i_type;
				-- cpl_o <= null_cpl_i_type;
				-- cpld_o <= null_cpld_i_type;

				mem_rd_o_ack_buf <= '0';
				mem_wr_o_ack_buf <= '0';
				cpld_o_ack_buf <= '0';
				cpl_o_ack_buf <= '0';

				tx_mem_rd_o <= null_mem_rd_o_type;
				tx_mem_wr_o <= null_mem_wr_o_type;
				tx_cpld_o <= null_cpld_o_type;

				transformation_state <= rst;

				general_busy <= '0';

				incoming_read_count <= (others => '0');
				incoming_write_count <= (others => '0');
				incoming_cpld_count <= (others => '0');
				outgoing_read_count <= (others => '0');
				outgoing_write_count <= (others => '0');
			else
				if (app_request_i.ack = '1' and app_request_o_reg.stb = '1' and app_request_o_reg.cyc = '1' and transformation_state /= rst) then
					-- successful termination of a WB cycle, remember the ack for later
					app_ack_received := '1';

					app_request_o_reg.stb <= '0';
					app_request_o_reg.cyc <= '0';

					-- if this was a read request, store the data that was read
					if (app_request_o_reg.we = '0') then
						data := app_request_i.dat;
						relevant_dat <= app_request_i.dat;
					end if;
				end if;

				case (transformation_state) is
					when rst => -- reset stuff, the usual
						in_rst_state <= '1';
						in_out_read_state <= '0';
						in_out_write_state <= '0';
						in_in_cpld_state <= '0';
						in_out_cpld_state <= '0';
						in_wait_cpld_state <= '0';
						in_wait_a_bit_state <= '0';
						in_wait_for_ack_state <= '0';

						write_enable := '0';
						data := (others => '0');
						address := (others => '0');
						transaction_counter := (others => '0');

						general_busy <= '0';
						cpld_o.busy <= '0'; -- should this be '1'?

						mem_request_o_reg.ack <= '0';

						app_ack_received := '0';

						-- handle all config BAR (BAR0) requests immediately
						-- we rely on BAR0 to be 1024 bytes at most (10 addr bits)
						if (mem_rd_rbar_hit_i(0) = '1' and mem_rd_i.valid = '1') then
							mem_tlp := mem_rd_i.tlp;
							case (to_integer(unsigned(mem_tlp.addr(9 downto 0)))) is
								when 1 => tx_cpld_o.data <= dma_base_address_reg;
								when 2 => tx_cpld_o.data <= dma_region_size;
								when 3 => tx_cpld_o.data <= xor_pattern;
								when 8 => tx_cpld_o.data <= incoming_read_count; -- tlp counters, readonly
								when 9 => tx_cpld_o.data <= incoming_write_count;
								when 10 => tx_cpld_o.data <= incoming_cpld_count;
								when 11 => tx_cpld_o.data <= outgoing_read_count;
								when 12 => tx_cpld_o.data <= outgoing_write_count;
								when 13 => tx_cpld_o.data <= outgoing_cpld_count;
								when others => tx_cpld_o.data <= X"AAAAAAAA";
							end case;
							tx_cpld_o.tlp.requester_id <= mem_tlp.requester_id;
							tx_cpld_o.tlp.tag <= mem_tlp.tag;
							tx_cpld_o.tlp.lower_addr <= mem_tlp.addr(6 downto 0);
							tx_cpld_o.tlp.byte_count <= (others => '0');
							tx_cpld_o.tlp.cpl_status <= cpl_status_successful_cpl;
							tx_cpld_o.tlp.common.length <= "0000000001";
							tx_cpld_o.valid <= '1';
							tx_cpld_o.data_valid <= '1';

							mem_rd_o_ack_buf <= '1';

							app_request_o_reg.stb <= '0';
							app_request_o_reg.cyc <= '0';
							app_request_o_reg.we <= '0';

							transformation_state <= rst;
						elsif (mem_wr_rbar_hit_i(0) = '1' and mem_wr_i.valid = '1' and mem_wr_i.data_valid = '1') then
							case (to_integer(unsigned(mem_wr_i.tlp.addr(9 downto 0)))) is
								when 1 => dma_base_address_reg <= bswap32(mem_wr_i.data);
								when 2 => dma_region_size <= bswap32(mem_wr_i.data);
								when 3 => xor_pattern <= bswap32(mem_wr_i.data);
								when others => null; -- ignore writes anywhere else
							end case;

							mem_wr_o_ack_buf <= '1';

							app_request_o_reg.stb <= '0';
							app_request_o_reg.cyc <= '0';
							app_request_o_reg.we <= '0';

							transformation_state <= rst;
						-- incoming read, CPU wants to read stuff from us
						elsif (mem_rd_rbar_hit_i(1) = '1' and mem_rd_i.valid = '1') then
							-- we have an incoming read on BAR1
							transformation_state <= wait_a_bit;
							write_enable := '0';
							general_busy <= '1';

							app_request_o_reg.we <= '0';
							app_request_o_reg.stb <= '1';
							app_request_o_reg.cyc <= '1';

							app_request_o_reg.adr <= mem_rd_i.tlp.addr & "00";
							relevant_adr <= mem_rd_i.tlp.addr & "00";

							-- we need to remember this to create the completion
							original_requester_id := mem_rd_i.tlp.requester_id;
							original_tag := mem_rd_i.tlp.tag;
							original_lower_addr := mem_rd_i.tlp.addr(4 downto 0);

							incoming_read_count <= incr(incoming_read_count);

							mem_rd_o_ack_buf <= '1';

						-- incoming write, CPU wants to write stuff
						elsif (mem_wr_rbar_hit_i(1) = '1' and mem_wr_i.valid = '1' and mem_wr_i.data_valid = '1') then
							-- we have incoming write on BAR1
							transformation_state <= wait_a_bit;
							write_enable := '1';
							general_busy <= '1';

							app_request_o_reg.we <= '1';
							app_request_o_reg.stb <= '1';
							app_request_o_reg.cyc <= '1';

							app_request_o_reg.adr <= mem_wr_i.tlp.addr & "00";
							app_request_o_reg.dat <= mem_wr_i.data;
							relevant_dat <= mem_wr_i.data;
							relevant_adr <= mem_wr_i.tlp.addr & "00";

							incoming_write_count <= incr(incoming_write_count);

							mem_wr_o_ack_buf <= '1';
						else
							app_request_o_reg.stb <= '0';
							app_request_o_reg.cyc <= '0';
							app_request_o_reg.we <= '0';

							transformation_state <= rst;
						end if;

						-- react to all acks (while still) in rst state
						if (tx_cpld_i.ack = '1') then
							tx_cpld_o.data_valid <= '0';
							tx_cpld_o.valid <= '0';
						end if;
						if (tx_mem_rd_i.ack = '1') then
							tx_mem_rd_o.valid <= '0';
						end if;
						if (tx_mem_wr_i.ack = '1') then
							tx_mem_wr_o.valid <= '0';
							tx_mem_wr_o.data_valid <= '0';
						end if;

					when wait_for_ack =>
						in_rst_state <= '0';
						in_out_read_state <= '0';
						in_out_write_state <= '0';
						in_in_cpld_state <= '0';
						in_out_cpld_state <= '0';
						in_wait_cpld_state <= '0';
						in_wait_a_bit_state <= '0';
						in_wait_for_ack_state <= '1';

						general_busy <= '1';
						cpld_o.busy <= '1';

						-- wait_for_cpld should have acked the mem_request cycle, so we just clean up here
						mem_request_o_reg.ack <= '0';
						mem_ack_sent := '0';

						-- some earlier state already received an ack or we are getting an ack just now
						if (app_ack_received = '1') then
							-- the app has data for us
							if (write_enable = '1') then
								-- we got a write from the CPU, so there will not be
								-- a completion
								transformation_state <= rst;
							else 
								outgoing_cpld_count <= incr(outgoing_cpld_count);
								transformation_state <= outgoing_cpld;
							end if;
						else
							transformation_state <= wait_a_bit;
						end if;

					when wait_a_bit => -- after, wait for application to finish and handle
						           -- request from the application

						-- clear acks set by outgoing_write and wait_cpld
						in_rst_state <= '0';
						in_out_read_state <= '0';
						in_out_write_state <= '0';
						in_in_cpld_state <= '0';
						in_out_cpld_state <= '0';
						in_wait_cpld_state <= '0';
						in_wait_a_bit_state <= '1';
						in_wait_for_ack_state <= '0';

						if (tx_mem_wr_i.ack = '1') then
							tx_mem_wr_o.valid <= '0';
							tx_mem_wr_o.data_valid <= '0';
						end if;

						general_busy <= '1';
						cpld_o.busy <= '1';

						if (app_ack_received = '1') then
							mem_request_o_reg.ack <= '0';
							mem_ack_sent := '0';

							-- the app has data for us
							if (write_enable = '1') then
								-- we got a write from the CPU, so there will not be
								-- a completion
								transformation_state <= rst;
							else 
								-- we got a read, so we have to relay the data via
								-- a completion
								data := app_request_i.dat;
								relevant_dat <= app_request_i.dat;

								outgoing_cpld_count <= incr(outgoing_cpld_count);
								transformation_state <= outgoing_cpld;
							end if;
						elsif (mem_request_i.cyc = '1' and mem_request_i.stb = '1' and mem_request_i.we = '1') then
							-- since there is no real error handling or possibility of error, we ack this
							-- as early as possible
							mem_request_o_reg.ack <= '1';
							mem_ack_sent := '1';
							-- the application wants to write to DMA memory
							transformation_state <= outgoing_write;

							-- get it while its still hot, stuff is only good during stb = '1'
							data := mem_request_i.dat;
							address := mem_request_i.adr;
							relevant_dat <= mem_request_i.dat;
							relevant_adr <= mem_request_i.adr;

							outgoing_write_count <= incr(outgoing_write_count);
						elsif (mem_request_i.cyc = '1' and mem_request_i.stb = '1' and mem_request_i.we = '0') then
							-- the application wants to read from DMA memory
							mem_request_o_reg.ack <= '0';
							mem_ack_sent := '0';

							transformation_state <= outgoing_read;
							address := mem_request_i.adr;
							relevant_adr <= mem_request_i.adr;

							outgoing_read_count <= incr(outgoing_read_count);
						else
							mem_request_o_reg.ack <= '0';
							mem_ack_sent := '0';
						end if;

					when outgoing_read =>
						in_rst_state <= '0';
						in_out_read_state <= '1';
						in_out_write_state <= '0';
						in_in_cpld_state <= '0';
						in_out_cpld_state <= '0';
						in_wait_cpld_state <= '0';
						in_wait_a_bit_state <= '0';
						in_wait_for_ack_state <= '0';

						general_busy <= '1';
						cpld_o.busy <= '1';

						tx_mem_rd_o.tlp.addr <= address(31 downto 2);
						tx_mem_rd_o.tlp.first_byte_enable <= (others => '1');
						tx_mem_rd_o.tlp.last_byte_enable <= (others => '1');
						tx_mem_rd_o.tlp.tag <= "01" & transaction_counter;
						tx_mem_rd_o.tlp.common.length <= "0000000001"; -- 4 bytes

						if (tx_mem_rd_i.ack = '0') then 
							tx_mem_rd_o.valid <= '1';
							transformation_state <= outgoing_read;
						else
							tx_mem_rd_o.valid <= '0';
							transformation_state <= wait_cpld;
						end if;

					when wait_cpld =>
						in_rst_state <= '0';
						in_out_read_state <= '0';
						in_out_write_state <= '0';
						in_in_cpld_state <= '0';
						in_out_cpld_state <= '0';
						in_wait_cpld_state <= '1';
						in_wait_a_bit_state <= '0';
						in_wait_for_ack_state <= '0';

						if (tx_mem_rd_i.ack = '1') then
							tx_mem_rd_o.valid <= '0';
						end if;

						general_busy <= '1';
						cpld_o.busy <= '0'; -- this instructs tlpgen to request reordering of TLPs so we get CPL(D)s first (hopefully)

						-- wait for a completion matching the read TLP we sent
						if (cpld_i.valid = '1' and cpld_i.data_valid = '1' and cpld_i.tlp.tag = "01" & transaction_counter) then
							transaction_counter := incr(transaction_counter);

							cpld_o_ack_buf <= '1';

							mem_request_o_reg.dat <= cpld_i.data;

							mem_ack_sent := '1';
							mem_request_o_reg.ack <= '1';

							transformation_state <= wait_for_ack;
							relevant_dat <= cpld_i.data;

							incoming_cpld_count <= incr(incoming_cpld_count);
						else
							mem_request_o_reg.ack <= '0';
							transformation_state <= wait_cpld;
						end if;

					when outgoing_write =>
						in_rst_state <= '0';
						in_out_read_state <= '0';
						in_out_write_state <= '1';
						in_in_cpld_state <= '0';
						in_out_cpld_state <= '0';
						in_wait_cpld_state <= '0';
						in_wait_a_bit_state <= '0';
						in_wait_for_ack_state <= '0';

						general_busy <= '1';
						cpld_o.busy <= '1';

						tx_mem_wr_o.tlp.addr <= address(31 downto 2);
						tx_mem_wr_o.data <= data;
						relevant_dat <= data;
						relevant_adr <= address;
						tx_mem_wr_o.tlp.common.length <= "0000000001"; -- 4 bytes
						tx_mem_wr_o.tlp.tag <= "01" & transaction_counter;
						tx_mem_wr_o.tlp.first_byte_enable <= (others => '1');
						tx_mem_wr_o.tlp.last_byte_enable <= (others => '1');

						-- only enable .ack for 1 cycle
						if (mem_ack_sent = '1') then
							mem_request_o_reg.ack <= '0';
						else
							mem_request_o_reg.ack <= '1';
							mem_ack_sent := '1';
						end if;

						if (tx_mem_wr_i.ack = '0') then
							tx_mem_wr_o.valid <= '1';
							tx_mem_wr_o.data_valid <= '1';
							transformation_state <= outgoing_write;
						else
							tx_mem_wr_o.valid <= '0';
							tx_mem_wr_o.data_valid <= '0';
							transformation_state <= wait_a_bit;
							transaction_counter := incr(transaction_counter);
						end if;

					when outgoing_cpld =>
						in_rst_state <= '0';
						in_out_read_state <= '0';
						in_out_write_state <= '0';
						in_in_cpld_state <= '0';
						in_out_cpld_state <= '1';
						in_wait_cpld_state <= '0';
						in_wait_a_bit_state <= '0';

						general_busy <= '1';
						cpld_o.busy <= '1';

						tx_cpld_o.data <= data;
						tx_cpld_o.tlp.common.length <= "0000000001"; -- 4 bytes
						tx_cpld_o.tlp.byte_count <=  "000000000100"; -- also 4 bytes
						tx_cpld_o.tlp.lower_addr <= original_lower_addr & "00"; -- see PCIe base spec. ch. 2.3.1.1
						tx_cpld_o.tlp.requester_id <= original_requester_id;
						tx_cpld_o.tlp.tag <= original_tag;
						tx_cpld_o.tlp.cpl_status <= cpl_status_successful_cpl;

						relevant_dat <= data;

						if (tx_cpld_i.ack = '0') then
							tx_cpld_o.valid <= '1';
							tx_cpld_o.data_valid <= '1';
							transformation_state <= outgoing_cpld;
						else
							tx_cpld_o.valid <= '0';
							tx_cpld_o.data_valid <= '0';
							app_request_o_reg.cyc <= '0';

							transformation_state <= rst;
						end if;

					when others =>
						transformation_state <= rst;
				end case;

				-- in addition to the big state machine stuff, make sure that
				-- acks are always cleared if the have served their purpose, i.e.
				-- their corresponding .valid has been cleared by tlpgen
				if (mem_wr_i.valid = '0' and mem_wr_o_ack_buf = '1') then
					mem_wr_o_ack_buf <= '0';
				end if;
				if (mem_rd_i.valid = '0' and mem_rd_o_ack_buf = '1') then
					mem_rd_o_ack_buf <= '0';
				end if;
				if (cpld_i.valid = '0' and cpld_o_ack_buf = '1') then
					cpld_o_ack_buf <= '0';
				end if;
				if (cpl_i.valid = '0' and cpl_o_ack_buf = '1') then
					cpl_o_ack_buf <= '0';
				end if;
			end if;
		end if;
	end process handle_data_transformation;

end architecture exzess_pcie_app_if_arch;
