library ieee;
use ieee.std_logic_1164.all;

library unisim;
use unisim.vcomponents.all;

library work;
use work.types.all;
use work.pcie.all;
use work.chipscope_h.all;

entity exzess_pcie_tlpgen is
	port (
		trn_i: in trn_i_type;
		trn_o: out trn_o_type;
		cfg_i: in cfg_i_type;
		cfg_o: out cfg_o_type;

		mem_rd_o: out mem_rd_o_type;
		mem_rd_rbar_hit_o: out std_logic_vector(6 downto 0);
		mem_rd_i: in mem_rd_i_type;

		mem_wr_o: out mem_wr_o_type;
		mem_wr_rbar_hit_o: out std_logic_vector(6 downto 0);
		mem_wr_i: in mem_wr_i_type;

		cpl_o: out cpl_o_type;
		cpl_i: in cpl_i_type;
		cpld_o: out cpld_o_type;
		cpld_i: in cpld_i_type;

		tx_mem_rd_i: in mem_rd_o_type; -- for the transmit direction, data type names are _reversed_
		tx_mem_rd_o: out mem_rd_i_type;

		tx_mem_wr_i: in mem_wr_o_type;
		tx_mem_wr_o: out mem_wr_i_type;

		tx_cpld_i: in cpld_o_type;
		tx_cpld_o: out cpld_i_type;

		diag_led: out std_logic;

		trn_clk: in std_logic;
		trn_reset_n: in std_logic--;
	);
end entity exzess_pcie_tlpgen;

architecture exzess_pcie_tlpgen_arch of exzess_pcie_tlpgen is
	-- dw0 is implicitly read in rst state since the 0th dw always looks about the same
	type receive_state_type is (
		rst, -- reset
		mem_rd_dw1, mem_rd_dw2, mem_rd_wt, -- memory read
		mem_wr_dw1, mem_wr_dw2, mem_wr_dw3, mem_wr_wt, -- memory write
		cpl_dw1, cpl_dw2, cpl_wt,-- completion without data
		cpld_dw1, cpld_dw2, cpld_dwn, cpld_wt -- completion with data
	);

	-- One more state dw0 each for transmit, to avoid exzessively ugly code when assembling
	-- the 0th TLP header dw.
	-- Of course this costs one cycle on each transmission.
	type transmit_state_type is (
		rst, -- reset
		cpld_dw0, cpld_dw1, cpld_dw2, cpld_dw3, cpld_wt, -- completion data
		mem_wr_dw0, mem_wr_dw1, mem_wr_dw2, mem_wr_dwn, -- memory write
		mem_rd_dw0, mem_rd_dw1, mem_rd_dw2
	);

	attribute fsm_encoding: string;

	signal transmit_state: transmit_state_type;
	attribute fsm_encoding of transmit_state: signal is "one-hot";

	signal receive_state: receive_state_type;
	attribute fsm_encoding of receive_state: signal is "one-hot";

	-- to make VHDL happy: buffer the trn_o output so we can read from our own output
	-- shared variable trn_o_buf: trn_o_type;
	signal rdst_rdy: std_logic;

	signal cs_control0: std_logic_vector(35 downto 0);
	signal cs_trig0: std_logic_vector(127 downto 0);

	signal trn_o_buf: trn_o_type;

	signal trigger: std_logic;
	signal busy_transmitting: std_logic;

	signal tx_mem_wr_o_ack_buf: std_logic;
	signal tx_mem_rd_o_ack_buf: std_logic;
	signal tx_cpld_o_ack_buf: std_logic;
begin
	-- static assignments

	trn_o <= trn_o_buf;

	trn_o_buf.tx.terrfwd <= '0';
	trn_o_buf.tx.tstr <= '0';
	trn_o_buf.tx.tcfg_gnt <= '1';

	trn_o_buf.fc.sel <= (others => '0');

	tx_cpld_o.ack <= tx_cpld_o_ack_buf;
	tx_mem_wr_o.ack <= tx_mem_wr_o_ack_buf;
	tx_mem_rd_o.ack <= tx_mem_rd_o_ack_buf;

	-- cs_icon: chipscope_icon port map (control0 => cs_control0);
	-- cs_ila: chipscope_ila port map (control => cs_control0, clk => trn_clk, trig0 => cs_trig0);

	cs_trig0(0) <= trn_reset_n;
	cs_trig0(1) <= trn_i.rx.rsrc_rdy;
	cs_trig0(2) <= rdst_rdy;
	cs_trig0(3) <= trn_i.rx.rsof;
	-- cs_trig0(4) <= mem_rd_o.valid; (out port, see below)
	cs_trig0(5) <= mem_rd_i.busy;
	cs_trig0(6) <= trigger;
	cs_trig0(41 downto 10) <= trn_i.rx.rd;
	cs_trig0(56 downto 50) <= trn_i.rx.rbar_hit;
	cs_trig0(42) <= trn_i.tx.tdst_rdy;
	cs_trig0(43) <= tx_mem_wr_i.valid;
	cs_trig0(44) <= tx_mem_rd_i.valid;
	cs_trig0(46) <= tx_mem_wr_i.data_valid;
	cs_trig0(57) <= tx_cpld_i.valid;
	cs_trig0(58) <= tx_cpld_i.data_valid;
	-- cs_trig0(59) <= mem_wr_o.valid; (out port, see below)

	cs_trig0(7) <= trn_o_buf.tx.tsof; -- tsof
	cs_trig0(8) <= trn_o_buf.tx.teof; -- teof
	cs_trig0(9) <= trn_o_buf.tx.tsrc_rdy; -- tsrc_rdy
	cs_trig0(91 downto 60) <= trn_o_buf.tx.td;

	cs_trig0(92) <= tx_mem_rd_o_ack_buf;
	cs_trig0(93) <= tx_mem_wr_o_ack_buf;
	cs_trig0(94) <= tx_cpld_o_ack_buf;

	cs_trig0(95) <= busy_transmitting;

	-- cs_trig0(123 downto 101) <= unused

	cs_trig0(124) <= mem_rd_i.ack;
	cs_trig0(125) <= mem_wr_i.ack;
	cs_trig0(126) <= cpld_i.ack;
	-- cs_trig0(127) <= cpld_o.valid; (out port, see below)

	-- receive state machine
	receive_sm: process (trn_clk, trn_reset_n, receive_state, rdst_rdy, trn_i) is
		variable last_tlp_hdr: tlp_hdr_common;
		variable mem_rd_o_buf: mem_rd_o_type;
		variable mem_wr_o_buf: mem_wr_o_type;
		variable cpl_o_buf: cpl_o_type;
		variable cpld_o_buf: cpld_o_type;
		variable countdown: integer;
	begin
	if (trn_clk = '1' and trn_clk'event) then
		if (trn_reset_n = '0') then
			cs_trig0(100 downto 96) <= (others => '0');
			receive_state <= rst;
			mem_rd_o_buf := null_mem_rd_o_type;
			mem_wr_o_buf := null_mem_wr_o_type;
			cpl_o_buf := null_cpl_o_type;
			cpld_o_buf := null_cpld_o_type;

			-- these get read on reset, when the 0th DW is received
			-- as soon as we know which state we end up in, they get assigned to the corr. output
			last_tlp_hdr := decode_common_hdr(trn_i.rx.rd);
			rdst_rdy <= '1';
			trn_o_buf.rx.rdst_rdy <= '1';
			trn_o_buf.rx.rnp_ok <= '1';
			countdown := 0;
		else
			case (receive_state) is
				when rst =>
					cs_trig0(96) <= '1'; -- rst state for chipscope
					cs_trig0(100 downto 97) <= (others => '0');

					mem_rd_o_buf := null_mem_rd_o_type;
					mem_wr_o_buf := null_mem_wr_o_type;
					cpl_o_buf := null_cpl_o_type;
					cpld_o_buf := null_cpld_o_type;

					-- these get read on reset, when the 0th DW is received
					-- as soon as we know which state we end up in, they get assigned to the corr. output
					last_tlp_hdr := decode_common_hdr(trn_i.rx.rd);

					-- signal that we are busy if any of the receivers is
					-- but only in rst state, if we start receiving, always receive the whole TLP
					-- at once.
					-- also, being busy has a timeout so we don't hang up the machine by never
					-- processing anything.
					if (countdown > 255) then
						-- we have waited long enough, something must be wrong...
						rdst_rdy <= '1';
						trn_o_buf.rx.rdst_rdy <= '1';
						trn_o_buf.rx.rnp_ok <= '1';
					elsif (countdown < 0) then
						-- this should never happen...
						countdown := 0;

						rdst_rdy <= '0';
						trn_o_buf.rx.rdst_rdy <= '0';
					elsif (countdown < 4) then
						-- do not receive the next TLP immediately, add some delay
						countdown := countdown + 1;
						rdst_rdy <= '0';
						trn_o_buf.rx.rdst_rdy <= '0';
						trn_o_buf.rx.rnp_ok <= '0';
					elsif ((mem_rd_i.busy = '1' or mem_wr_i.busy = '1') and cpld_i.busy = '1') then
						-- cpld_i says busy, dont receive anything
						countdown := countdown + 1;
						rdst_rdy <= '0';
						trn_o_buf.rx.rdst_rdy <= '0';
						trn_o_buf.rx.rnp_ok <= '0';
					elsif ((mem_rd_i.busy = '1' or mem_wr_i.busy = '1') and cpld_i.busy = '0') then
						-- cpld_i says not busy, receive, but only cplds
						countdown := countdown + 1;
						rdst_rdy <= '1';
						trn_o_buf.rx.rdst_rdy <= '1';
						trn_o_buf.rx.rnp_ok <= '0';
					else
						-- nothing is busy, receive stuff
						countdown := countdown + 1;
						trn_o_buf.rx.rnp_ok <= '1';
						rdst_rdy <= '1';
						trn_o_buf.rx.rdst_rdy <= '1';
					end if;
					-- trn_o_buf.rx.rdst_rdy <= (not mem_rd_i.busy) or (not mem_wr_i.busy) or (not cpld_i.busy);
					-- rdst_rdy <= '1';
					-- trn_o_buf.rx.rdst_rdy <= '1';

					--  start of frame, source ready, destination ready
					if (trn_i.rx.rsof = '1' and trn_i.rx.rsrc_rdy = '1' and rdst_rdy = '1' and countdown > 2) then
						trigger <= '1'; -- just to tell chipscope that interesting stuff is happening
						countdown := 0;
						case (trn_i.rx.rd(30 downto 24)) is
							when mem_rd32_magic =>
								receive_state <= mem_rd_dw1;
							when mem_wr32_magic =>
								receive_state <= mem_wr_dw1;
							when cpl_magic =>
								receive_state <= cpl_dw1;
							when cpld_magic =>
								receive_state <= cpld_dw1;
							when others =>
								receive_state <= rst;
						end case;
					else
						trigger <= '0';
						receive_state <= rst;
					end if;

				when mem_rd_dw1 =>
					cs_trig0(97) <= '1'; -- mem_rd state for chipscope
					cs_trig0(96) <= '0';
					mem_rd_o_buf.tlp.common := last_tlp_hdr;
					decode_mem_dw1(trn_i.rx.rd, mem_rd_o_buf.tlp);

					if (trn_i.rx.rsrc_rdy = '1') then
						receive_state <= mem_rd_dw2;
					else
						receive_state <= mem_rd_dw1;
					end if;

				when mem_rd_dw2 =>
					decode_mem_dw2(trn_i.rx.rd, mem_rd_o_buf.tlp);
					mem_rd_o_buf.valid := '1';
					mem_rd_o <= mem_rd_o_buf;
					mem_rd_rbar_hit_o <= trn_i.rx.rbar_hit;
					cs_trig0(4) <= '1';

					--if (trn_i.rx.reof = '1' and trn_i.rx.rsrc_rdy = '1') then
					-- eof is useless here and might lead to hangs
					if (trn_i.rx.rsrc_rdy = '1') then
						trn_o_buf.rx.rdst_rdy <= '0';
						rdst_rdy <= '0';
						receive_state <= mem_rd_wt;
					else
						receive_state <= mem_rd_dw2;
					end if;

				when mem_rd_wt =>
					trn_o_buf.rx.rdst_rdy <= '0';
					mem_rd_o.valid <= '1';
					rdst_rdy <= '0';
					receive_state <= rst;

				when mem_wr_dw1 =>
					cs_trig0(98) <= '1'; -- mem_rd state for chipscope
					cs_trig0(96) <= '0';
					mem_wr_o_buf.tlp.common := last_tlp_hdr;
					decode_mem_dw1(trn_i.rx.rd, mem_wr_o_buf.tlp);

					if (trn_i.rx.rsrc_rdy = '1') then
						receive_state <= mem_wr_dw2;
					else
						receive_state <= mem_wr_dw1;
					end if;

				when mem_wr_dw2 =>
					decode_mem_dw2(trn_i.rx.rd, mem_wr_o_buf.tlp);
					mem_wr_o_buf.valid := '1';
					mem_wr_o <= mem_wr_o_buf;
					mem_wr_rbar_hit_o <= trn_i.rx.rbar_hit;
					cs_trig0(59) <= '1';

					if (trn_i.rx.rsrc_rdy = '1') then
						receive_state <= mem_wr_dw3;
					else
						receive_state <= mem_wr_dw2;
					end if;

				when mem_wr_dw3 =>
					mem_wr_o.data <= trn_i.rx.rd;
					mem_wr_o.valid <= '1';

					if (trn_i.rx.rsrc_rdy = '1') then
						trn_o_buf.rx.rdst_rdy <= '0';
						rdst_rdy <= '0';
						receive_state <= mem_wr_wt;
					else
						receive_state <= mem_wr_dw3;
					end if;

				when mem_wr_wt =>
					mem_wr_o.data_valid <= '1';
					trn_o_buf.rx.rdst_rdy <= '0';
					rdst_rdy <= '0';
					receive_state <= rst; -- FIXME: perhaps we need to wait?

				when cpl_dw1 =>
					cs_trig0(99) <= '1'; -- mem_rd state for chipscope
					cs_trig0(96) <= '0';
					cpl_o_buf.tlp.common := last_tlp_hdr;
					decode_cpl_dw1(trn_i.rx.rd, cpl_o_buf.tlp);

					if (trn_i.rx.rsrc_rdy = '1') then
						receive_state <= cpl_dw2;
					else
						receive_state <= cpl_dw1;
					end if;

				when cpl_dw2 =>
					decode_cpl_dw2(trn_i.rx.rd, cpl_o_buf.tlp);
					cpl_o <= cpl_o_buf;

					if (trn_i.rx.rsrc_rdy = '1') then
						receive_state <= cpl_wt;
						trn_o_buf.rx.rdst_rdy <= '0';
						rdst_rdy <= '0';
					else
						receive_state <= cpl_dw2;
					end if;
				when cpl_wt =>
					cpl_o.valid <= '1';
					trn_o_buf.rx.rdst_rdy <= '0';
					rdst_rdy <= '0';
					receive_state <= rst;

				when cpld_dw1 =>
					cs_trig0(100) <= '1'; -- mem_rd state for chipscope
					cs_trig0(96) <= '0';
					cpld_o_buf.tlp.common := last_tlp_hdr;
					decode_cpl_dw1(trn_i.rx.rd, cpld_o_buf.tlp);

					if (trn_i.rx.rsrc_rdy = '1') then
						receive_state <= cpld_dw2;
					else
						receive_state <= cpld_dw1;
					end if;

				when cpld_dw2 =>
					decode_cpl_dw2(trn_i.rx.rd, cpld_o_buf.tlp);
					cpld_o_buf.valid := '1';
					cpld_o <= cpld_o_buf;
					cs_trig0(127) <= '1';

					if (trn_i.rx.rsrc_rdy = '1') then
						receive_state <= cpld_dwn;
					else
						receive_state <= cpld_dw2;
					end if;

				when cpld_dwn =>
					cpld_o.data <= trn_i.rx.rd;
					cpld_o.valid <= '1';

					if (trn_i.rx.rsrc_rdy = '1') then
						receive_state <= cpld_wt;
						trn_o_buf.rx.rdst_rdy <= '0';
						rdst_rdy <= '0';
					else
						receive_state <= cpld_dwn;
					end if;
				when cpld_wt =>
					cpld_o.data_valid <= '1';
					trn_o_buf.rx.rdst_rdy <= '0';
					rdst_rdy <= '0';
					receive_state <= rst;

				when others =>
					receive_state <= rst;
			end case;

			-- reset valid bits if reception of data is ack'ed
			if (mem_rd_i.ack = '1') then
				mem_rd_o.valid <= '0';
				cs_trig0(4) <= '0';
			end if;
			if (mem_wr_i.ack = '1') then
				mem_wr_o.valid <= '0';
				cs_trig0(59) <= '0';
			end if;
			if (cpl_i.ack = '1') then
				cpl_o.valid <= '0';
			end if;
			if (cpld_i.ack = '1') then
				cpld_o.valid <= '0';
				cs_trig0(127) <= '0';
			end if;

		end if;
	end if;
	end process receive_sm;

	-- we are busy transmitting...
	-- doesn't matter what we are transmitting, so split up the
	-- busy signal to all (possibly) interested outputs
	tx_mem_rd_o.busy <= busy_transmitting;
	tx_mem_wr_o.busy <= busy_transmitting;
	tx_cpld_o.busy <= busy_transmitting;

	transmit_sm: process (transmit_state, trn_i, tx_mem_rd_i, tx_mem_wr_i, tx_cpld_i, cfg_i, trn_reset_n, trn_clk) is
		variable tx_cpld_i_buf: cpld_o_type;
		variable tx_mem_rd_i_buf: mem_rd_o_type;
		variable tx_mem_wr_i_buf: mem_wr_o_type;
	begin
	if (trn_clk = '1' and trn_clk'event) then
		if (trn_reset_n = '0') then
			transmit_state <= rst;
		else
			case (transmit_state) is
				when rst =>
					trn_o_buf.tx.tsrc_rdy <= '0';
					if (trn_i.tx.tdst_rdy = '1') then
						-- the 'if' makes sure that stuff stays asserted until the destination has
						-- read the last data (I guess)
						trn_o_buf.tx.tsof <= '0';
						trn_o_buf.tx.teof <= '0';
						trn_o_buf.tx.tsrc_rdy <= '0';
						trn_o_buf.tx.tsrc_dsc <= '0';
						trn_o_buf.tx.td <= (others => '0');
						-- trn_i.tx.tdst_dsc = '0' is unnecessary since
						-- the xilinx xapp1052 always sets it to that
						-- value. no idea why they use that signal in the first
						-- place. symmetry? unnecessary with my version of the core?

						-- strictly priority based send, maybe I need to replace this by
						-- round robin to avoid starving one TLP sender type
						if (tx_cpld_i.valid = '1') then
							tx_cpld_o_ack_buf <= '1';
							tx_cpld_i_buf := tx_cpld_i;
							transmit_state <= cpld_dw0;
							busy_transmitting <= '1';
							cs_trig0(45) <= '1';
							cs_trig0(49) <= '0';
						elsif (tx_mem_rd_i.valid = '1') then
							tx_mem_rd_o_ack_buf <= '1';
							tx_mem_rd_i_buf := tx_mem_rd_i;
							transmit_state <= mem_rd_dw0;
							busy_transmitting <= '1';
							cs_trig0(47) <= '1';
							cs_trig0(49) <= '0';
						elsif (tx_mem_wr_i.valid = '1') then
							tx_mem_wr_o_ack_buf <= '1';
							tx_mem_wr_i_buf := tx_mem_wr_i;
							transmit_state <= mem_wr_dw0;
							busy_transmitting <= '1';
							cs_trig0(48) <= '1';
							cs_trig0(49) <= '0';
						else
							tx_cpld_i_buf := null_cpld_o_type;
							tx_mem_wr_i_buf := null_mem_wr_o_type;
							tx_mem_rd_i_buf := null_mem_rd_o_type;

							busy_transmitting <= '0';
							transmit_state <= rst;
							cs_trig0(49) <= '1'; -- signal state (1-hot) to chipscope
							cs_trig0(45) <= '0'; -- at least the current state branch
							cs_trig0(47) <= '0';
							cs_trig0(48) <= '0';
						end if;
					else
						busy_transmitting <= '0';
						transmit_state <= rst;

						cs_trig0(49) <= '1'; -- signal state (1-hot) to chipscope
						cs_trig0(45) <= '0'; -- at least the current state branch
						cs_trig0(47) <= '0';
						cs_trig0(48) <= '0';
					end if;
				when mem_rd_dw0 =>
					trn_o_buf.tx.tsof <= '1';
					trn_o_buf.tx.teof <= '0';
					trn_o_buf.tx.tsrc_rdy <= '1';

					tx_mem_rd_i_buf.tlp.requester_id := create_endpoint_id_from_cfg(cfg_i);
					trn_o_buf.tx.td <= assemble_common_hdr(mem_rd32_magic, tx_mem_rd_i_buf.tlp.common);

					if (trn_i.tx.tdst_rdy = '1') then
						transmit_state <= mem_rd_dw1;
					else
						transmit_state <= mem_rd_dw0;
					end if;

				when mem_rd_dw1 =>
					trn_o_buf.tx.tsof <= '0';
					trn_o_buf.tx.td <= assemble_mem_dw1(tx_mem_rd_i_buf.tlp);

					if (trn_i.tx.tdst_rdy = '1') then
						transmit_state <= mem_rd_dw2;
					else
						transmit_state <= mem_rd_dw1;
					end if;

				when mem_rd_dw2 =>
					trn_o_buf.tx.teof <= '1';
					trn_o_buf.tx.td <= assemble_mem_dw2(tx_mem_rd_i_buf.tlp);

					if (trn_i.tx.tdst_rdy = '1') then
						transmit_state <= rst;
					else
						transmit_state <= mem_rd_dw2;
					end if;

				when mem_wr_dw0 =>
					trn_o_buf.tx.tsof <= '1';
					trn_o_buf.tx.teof <= '0';
					trn_o_buf.tx.tsrc_rdy <= '1';

					tx_mem_wr_i_buf.tlp.requester_id := create_endpoint_id_from_cfg(cfg_i);
					trn_o_buf.tx.td <= assemble_common_hdr(mem_wr32_magic, tx_mem_wr_i_buf.tlp.common);
					if (trn_i.tx.tdst_rdy = '1') then
						transmit_state <= mem_wr_dw1;
					else
						transmit_state <= mem_wr_dw0;
					end if;

				when mem_wr_dw1 =>
					trn_o_buf.tx.tsof <= '0';
					trn_o_buf.tx.td <= assemble_mem_dw1(tx_mem_wr_i_buf.tlp);

					if (trn_i.tx.tdst_rdy = '1') then
						transmit_state <= mem_wr_dw2;
					else
						transmit_state <= mem_wr_dw1;
					end if;

				when mem_wr_dw2 =>
					trn_o_buf.tx.td <= assemble_mem_dw2(tx_mem_wr_i_buf.tlp);

					if (trn_i.tx.tdst_rdy = '1') then
						transmit_state <= mem_wr_dwn;
					else
						transmit_state <= mem_wr_dw2;
					end if;

				when mem_wr_dwn =>
					-- FIXME: implement transmission of more than one dw
					trn_o_buf.tx.teof <= '1';
					trn_o_buf.tx.td <= tx_mem_wr_i.data;

					-- FIXME: and count < max
					if (trn_i.tx.tdst_rdy = '1') then
						transmit_state <= rst;
					else
						transmit_state <= mem_wr_dwn;
					end if;

				when cpld_dw0 =>
					tx_cpld_i_buf.tlp.completer_id := create_endpoint_id_from_cfg(cfg_i);

					trn_o_buf.tx.tsof <= '1';
					trn_o_buf.tx.teof <= '0';
					trn_o_buf.tx.tsrc_rdy <= '1';
					trn_o_buf.tx.td <= assemble_common_hdr(cpld_magic, tx_cpld_i_buf.tlp.common);

					if (trn_i.tx.tdst_rdy = '1') then
						transmit_state <= cpld_dw1;
					else
						transmit_state <= cpld_dw0;
					end if;

				when cpld_dw1 =>
					trn_o_buf.tx.tsof <= '0';
					trn_o_buf.tx.td <= assemble_cpl_dw1(tx_cpld_i_buf.tlp);

					if (trn_i.tx.tdst_rdy = '1') then
						transmit_state <= cpld_dw2;
					else
						transmit_state <= cpld_dw1;
					end if;

				when cpld_dw2 =>
					trn_o_buf.tx.td <= assemble_cpl_dw2(tx_cpld_i_buf.tlp);

					if (trn_i.tx.tdst_rdy = '1') then
						transmit_state <= cpld_dw3;
					else
						transmit_state <= cpld_dw2;
					end if;

				when cpld_dw3 =>
					-- FIXME: currently, only one data dw is transmitted. extend.
					trn_o_buf.tx.teof <= '1';
					trn_o_buf.tx.td <= tx_cpld_i.data;

					if (trn_i.tx.tdst_rdy = '1') then
						transmit_state <= cpld_wt;
					else
						transmit_state <= cpld_dw3;
					end if;

				when cpld_wt =>
					if (trn_i.tx.tdst_rdy = '1') then
						trn_o_buf.tx.tsof <= '0';
						trn_o_buf.tx.teof <= '0';
						trn_o_buf.tx.tsrc_rdy <= '0';
						trn_o_buf.tx.tsrc_dsc <= '0';
					end if;

					if (trn_i.tx.tdst_rdy = '1') then
						transmit_state <= rst;
					else
						transmit_state <= cpld_wt;
					end if;
				when others =>
					transmit_state <= rst;
			end case;

			-- reset acks
			if (tx_mem_wr_i.valid = '0' and tx_mem_wr_o_ack_buf = '1') then
				tx_mem_wr_o_ack_buf <= '0';
			end if;
			if (tx_mem_rd_i.valid = '0' and tx_mem_rd_o_ack_buf = '1') then
				tx_mem_rd_o_ack_buf <= '0';
			end if;
			if (tx_cpld_i.valid = '0' and tx_cpld_o_ack_buf = '1') then
				tx_cpld_o_ack_buf <= '0';
			end if;
		end if;
	end if;
	end process transmit_sm;

end architecture exzess_pcie_tlpgen_arch;
