library ieee;
use ieee.std_logic_1164.all;

library work;
use work.types.all;

package pcie is

-- magic numbers for PCIe TLP headers
-- from xilinx xapp1052, BMD_32_TX/RX_ENGINE.v
constant mem_rd32_magic: std_logic_vector(6 downto 0) := "0000000";
constant mem_wr32_magic: std_logic_vector(6 downto 0) := "1000000";
constant cpl_magic: std_logic_vector (6 downto 0) := "0001010";
constant cpld_magic: std_logic_vector (6 downto 0) := "1001010";

-- all headers in the first QW of a TLP
-- these headers are common to all types of TLPs
type tlp_hdr_common is record
	fmt: std_logic_vector(1 downto 0); -- TLP format: bit 1: data(1)/no data(0), bit 0: 3DW hdr(0)/4DW hdr(1)
	typ: std_logic_vector(4 downto 0); -- type of the request (see PCIe spec p. 47 for table)
	tc: std_logic_vector(2 downto 0); -- traffic class
	td: std_logic; -- tlp digest follows
	ep: std_logic; -- tlp is poisoned
	attr: std_logic_vector(1 downto 0); -- attributes: bit 1: relaxed ordering, bit 0: no snoop (== no cache coherency expected)
	length: std_logic_vector(9 downto 0); -- payload length (if there is any) or requested length of mem read
					      -- units are DWs. encoding for 1024 DWs is special (others => '0')
end record tlp_hdr_common;

-- needed because vhdl (others => '0') is useless for records...
constant null_tlp_hdr_common: tlp_hdr_common := (fmt => "00", typ => "00000", tc => "000", td => '0', ep => '0', attr => "00", length => (others => '0')); 

-- encapsulates a memory read or write tlp
type mem_tlp is record
	common: tlp_hdr_common;
	requester_id: std_logic_vector(15 downto 0);
	tag: std_logic_vector(7 downto 0);
	first_byte_enable: std_logic_vector(3 downto 0);
	last_byte_enable: std_logic_vector(3 downto 0);
	addr: std_logic_vector(29 downto 0);
end record mem_tlp;

constant null_mem_tlp: mem_tlp := (common => null_tlp_hdr_common, requester_id => (others => '0'), tag => (others => '0'), first_byte_enable => "0000", last_byte_enable => "0000", addr => (others => '0'));

type mem_rd_o_type is record
	tlp: mem_tlp;
	valid: std_logic;
end record mem_rd_o_type;
type mem_rd_i_type is record
	busy: std_logic;
	ack: std_logic;
end record mem_rd_i_type;

constant null_mem_rd_o_type: mem_rd_o_type := (tlp => null_mem_tlp, valid => '0');
constant null_mem_rd_i_type: mem_rd_i_type := (others => '0');

type mem_wr_o_type is record
	tlp: mem_tlp;
	data: std_logic_vector(31 downto 0);
	data_valid: std_logic;
	valid: std_logic;
end record mem_wr_o_type;

type mem_wr_i_type is record
	busy: std_logic;
	ack: std_logic;
end record mem_wr_i_type;

constant null_mem_wr_o_type: mem_wr_o_type := (tlp => null_mem_tlp, data => (others => '0'), data_valid => '0', valid => '0');
constant null_mem_wr_i_type: mem_wr_i_type := (others => '0');

type cpl_tlp is record
	common: tlp_hdr_common;
	completer_id: std_logic_vector(15 downto 0);
	cpl_status: std_logic_vector(2 downto 0); -- see corresponding constants
	bcm: std_logic; -- always zero for PCIe, may be nonzero for PCI-X
	byte_count: std_logic_vector(11 downto 0); -- remaining byte count for request
	requester_id: std_logic_vector(15 downto 0);
	tag: std_logic_vector(7 downto 0);
	lower_addr: std_logic_vector(6 downto 0); -- (lower part of) address this completion corresponds to
end record cpl_tlp;

constant null_cpl_tlp: cpl_tlp := (common => null_tlp_hdr_common, completer_id => (others => '0'), cpl_status => "000", bcm => '0', byte_count => (others => '0'), requester_id => (others => '0'), tag => (others => '0'), lower_addr => (others => '0'));

type cpl_o_type is record
	tlp: cpl_tlp;
	valid: std_logic;
end record cpl_o_type;

type cpl_i_type is record
	busy: std_logic;
	ack: std_logic;
end record cpl_i_type;

constant null_cpl_o_type: cpl_o_type := (tlp => null_cpl_tlp, valid => '0');
constant null_cpl_i_type: cpl_i_type := (others => '0');

type cpld_o_type is record
	tlp: cpl_tlp;
	data: std_logic_vector(31 downto 0);
	data_valid: std_logic;
	valid: std_logic;
end record cpld_o_type;

type cpld_i_type is record
	busy: std_logic;
	ack: std_logic;
end record cpld_i_type;

constant null_cpld_o_type: cpld_o_type := (tlp => null_cpl_tlp, data => (others => '0'), data_valid => '0', valid => '0');
constant null_cpld_i_type: cpld_i_type := (others => '0');

constant cpl_status_successful_cpl: std_logic_vector(2 downto 0) := "000";
constant cpl_status_unsupported_request: std_logic_vector(2 downto 0) := "001";
constant cpl_status_conf_req_retry_status: std_logic_vector(2 downto 0) := "010";
constant cpl_status_completer_abort: std_logic_vector(2 downto 0) := "100";

function decode_common_hdr(data: in std_logic_vector(31 downto 0)) return tlp_hdr_common;

procedure decode_mem_dw1(data: in std_logic_vector(31 downto 0); tlp: inout mem_tlp);
procedure decode_mem_dw2(data: in std_logic_vector(31 downto 0); tlp: inout mem_tlp);

procedure decode_cpl_dw1(data: in std_logic_vector(31 downto 0); tlp: inout cpl_tlp);
procedure decode_cpl_dw2(data: in std_logic_vector(31 downto 0); tlp: inout cpl_tlp);

function assemble_common_hdr(magic: in std_logic_vector(6 downto 0); common: in tlp_hdr_common) return std_logic_vector;
function assemble_mem_dw1(tlp: in mem_tlp) return std_logic_vector;
function assemble_mem_dw2(tlp: in mem_tlp) return std_logic_vector;
function assemble_cpl_dw1(tlp: in cpl_tlp) return std_logic_vector;
function assemble_cpl_dw2(tlp: in cpl_tlp) return std_logic_vector;

function create_endpoint_id(bus_id: in std_logic_vector(7 downto 0); dev_id: in std_logic_vector(4 downto 0); fn_id: in std_logic_vector(2 downto 0)) return std_logic_vector;
function create_endpoint_id_from_cfg(cfg: in cfg_i_type) return std_logic_vector;

end package pcie;

package body pcie is
	-- decode the common header (doubleword 0) of any TLP
	function decode_common_hdr(data: in std_logic_vector(31 downto 0)) return tlp_hdr_common is
		variable ret: tlp_hdr_common;
	begin
		ret.fmt := data(30 downto 29);
		ret.typ := data(28 downto 24);
		ret.tc := data(22 downto 20);
		ret.td := data(15);
		ret.ep := data(14);
		ret.attr := data(13 downto 12);
		ret.length := data(9 downto 0);

		return ret;
	end function decode_common_hdr;

	function assemble_common_hdr(magic: in std_logic_vector(6 downto 0); common: in tlp_hdr_common) return std_logic_vector is
		variable ret: std_logic_vector(31 downto 0);
	begin
		ret(31) := '0'; -- reserved
		ret(30 downto 24) := magic;
		ret(23) := '0'; -- reserved
		ret(22 downto 20) := common.tc;
		ret(19 downto 16) := (others => '0'); -- reserved
		ret(15) := common.td;
		ret(14) := '0'; -- common.ep, but nobody wants to create error poisoned TLPs I hope...
		ret(13 downto 12) := common.attr;
		ret(11 downto 10) := (others => '0'); -- reserved
		ret(9 downto 0) := common.length;

		return ret;
	end function assemble_common_hdr;

	-- decode the first DW of a memory (read/write) request TLP
	procedure decode_mem_dw1(data: in std_logic_vector(31 downto 0); tlp: inout mem_tlp) is
	begin
		tlp.requester_id := data(31 downto 16);
		tlp.tag := data(15 downto 8);
		tlp.last_byte_enable := data(7 downto 4);
		tlp.first_byte_enable := data(3 downto 0);
	end procedure decode_mem_dw1;

	function assemble_mem_dw1(tlp: in mem_tlp) return std_logic_vector is
		variable ret: std_logic_vector(31 downto 0);
	begin
		ret(31 downto 16) := tlp.requester_id;
		ret(15 downto 8) := tlp.tag;
		ret(7 downto 4) := tlp.last_byte_enable;
		ret(3 downto 0) := tlp.first_byte_enable;

		return ret;
	end function assemble_mem_dw1;

	procedure decode_mem_dw2(data: in std_logic_vector(31 downto 0); tlp: inout mem_tlp) is
	begin
		tlp.addr := data(31 downto 2);
	end procedure decode_mem_dw2;

	function assemble_mem_dw2(tlp: in mem_tlp) return std_logic_vector is
		variable ret: std_logic_vector(31 downto 0);
	begin
		ret(31 downto 2) := tlp.addr;
		ret(1 downto 0) := (others => '0'); -- reserved;

		return ret;
	end function assemble_mem_dw2;

	-- decode first DW of a completion (with or without data) TLP
	procedure decode_cpl_dw1(data: in std_logic_vector(31 downto 0); tlp: inout cpl_tlp) is
	begin
		tlp.completer_id := data(31 downto 16);
		tlp.cpl_status := data(15 downto 13);
		tlp.bcm := data(12);
		tlp.byte_count := data(11 downto 0);
	end procedure decode_cpl_dw1;

	function assemble_cpl_dw1(tlp: in cpl_tlp) return std_logic_vector is
		variable ret: std_logic_vector(31 downto 0);
	begin
		ret(31 downto 16) := tlp.completer_id;
		ret(15 downto 13) := tlp.cpl_status;
		ret(12) := '0'; -- tlp.bcm, may only be set by PCI-X completers, which we are not
		ret(11 downto 0) := tlp.byte_count;

		return ret;
	end function assemble_cpl_dw1;

	procedure decode_cpl_dw2(data: in std_logic_vector(31 downto 0); tlp: inout cpl_tlp) is
	begin
		tlp.requester_id := data(31 downto 16);
		tlp.tag := data(15 downto 8);
		tlp.lower_addr := data(6 downto 0);
	end procedure decode_cpl_dw2;			

	function assemble_cpl_dw2(tlp: in cpl_tlp) return std_logic_vector is
		variable ret: std_logic_vector(31 downto 0);
	begin
		ret(31 downto 16) := tlp.requester_id;
		ret(15 downto 8) := tlp.tag;
		ret(7) := '0'; -- reserved
		ret(6 downto 0) := tlp.lower_addr;

		return ret;
	end function assemble_cpl_dw2;

	-- create a requester or completer id
	function create_endpoint_id(bus_id: in std_logic_vector(7 downto 0); dev_id: in std_logic_vector(4 downto 0); fn_id: in std_logic_vector(2 downto 0)) return std_logic_vector is
		variable ret: std_logic_vector(15 downto 0);
	begin
		ret := bus_id & dev_id & fn_id;
		return ret;
	end function create_endpoint_id;

	function create_endpoint_id_from_cfg(cfg: in cfg_i_type) return std_logic_vector is
	begin
		return (create_endpoint_id(cfg.bus_number, cfg.device_number, cfg.function_number));
	end function create_endpoint_id_from_cfg;

end package body pcie;
