library ieee;
use ieee.std_logic_1164.all;

package types is

-- trn interface to pci core
-- flow control
type trn_fc_o_type is record
	sel: std_logic_vector (2 downto 0); -- flow control information select for all fc signals
	-- 000b: receive buffer available space
	-- 001b: receive credits granted to the link partner
	-- 010b: receive credits consumed
	-- 100b: transmit user credits available
	-- 101b: transmit credit limit
	-- 110b: transmit credits consumed
end record;

type trn_fc_i_type is record
	nph: std_logic_vector (7 downto 0); -- non-posted header
	npd: std_logic_vector (11 downto 0); -- non-posted data
	ph: std_logic_vector (7 downto 0); -- posted header
	pd: std_logic_vector (11 downto 0); -- posted data
	cplh: std_logic_vector (7 downto 0); -- completion header
	cpld: std_logic_vector (11 downto 0); -- completion data
end record;

-- tx
type trn_tx_o_type is record
	td: std_logic_vector (31 downto 0); -- tlp header and data dw
	tsof: std_logic; -- start of frame
	teof: std_logic; -- end of frame
	tsrc_rdy: std_logic; -- source ready
	tsrc_dsc: std_logic; -- discontinue, abort current tlp
	terrfwd: std_logic; -- transmit error forward, tlp is marked as error poisoned
	tstr: std_logic; -- transmit streamed, packet must be presented on consecutive clock cycles for
	                 -- immediate transmission (or terr_drop will occur)
	tcfg_gnt: std_logic; -- configuration grant, allow core to transmit internally generated conf tlps
end record;

type trn_tx_i_type is record
	tdst_rdy: std_logic; -- destination ready
	terr_drop: std_logic; -- transmit error drop, because of length or streaming timing violation
	tbuf_av: std_logic_vector (5 downto 0); -- num transmit buffers available
	tcfg_req: std_logic; -- configuration request, core requests to be allowed to transmit
	                     -- internally generated conf tlps
end record;

-- rx
type trn_rx_o_type is record
	rdst_rdy: std_logic; -- destination ready
	rnp_ok: std_logic; -- receive non-posted ok.
	-- Quote from Xilinx UG654 v3.0 2010-04-19:
	-- "Receive Non-Posted OK (active Low). The user application
	-- asserts this input whenever it is ready to accept a Non-Posted
	-- Request packet. This allows Posted and Completion packets to
	-- bypass Non-Posted packets in the inbound queue if necessitated
	-- by the user application. When the user application approaches a
	-- state where it is unable to service Non-Posted Requests, it must
	-- deassert TRNRNPOKN one clock cycle before the integrated
	-- Endpoint block presents TRNREOFN of the last Non-Posted TLP
	-- the user application can accept."

end record;

type trn_rx_i_type is record
	rd: std_logic_vector (31 downto 0);
	rsof: std_logic; -- start of frame
	reof: std_logic; -- end of frame
	rsrc_rdy: std_logic; -- source ready
	rsrc_dsc: std_logic; -- discontinue, abort current tlp
	rerrfwd: std_logic; -- receive error forward, currently received TLP is error poisoned
	rbar_hit: std_logic_vector (6 downto 0); -- BAR hit by current receive transaction
end record;

-- all trn subinterfaces combined
type trn_i_type is record
	fc: trn_fc_i_type;
	tx: trn_tx_i_type;
	rx: trn_rx_i_type;
	lnk_up: std_logic; -- PCIe link up, physical layer connection established
end record;

type trn_o_type is record
	fc: trn_fc_o_type;
	tx: trn_tx_o_type;
	rx: trn_rx_o_type;
end record;

-- config interface
type cfg_i_type is record
	do: std_logic_vector (31 downto 0); -- management data out, outputs configuration data from endpoint block
	rd_wr_done: std_logic; -- management read or write done
	err_cpl_rdy: std_logic; -- configuration error completion ready, indicates that endpoint core can accept err_ur and err_cpl_abort
	interrupt_rdy: std_logic; -- interrupt ready, if interrupt (see below) is high, interrupt has been transmitted
	interrupt_do: std_logic_vector (7 downto 0); -- interrupt data out
	interrupt_mmenable: std_logic_vector (2 downto 0); -- interrupt multiple message enabled
	interrupt_msienable: std_logic; -- MSI is enabled
	to_turnoff: std_logic; -- PME_TURN_OFF has been received, requesting us to turn off, waiting for turnoff_ok
	pcie_link_state: std_logic_vector(2 downto 0);
	bus_number: std_logic_vector(7 downto 0); 
	device_number: std_logic_vector(4 downto 0);
	function_number: std_logic_vector(2 downto 0);
	status: std_logic_vector(15 downto 0);
	command: std_logic_vector(15 downto 0);
	dstatus: std_logic_vector(15 downto 0);
	dcommand: std_logic_vector(15 downto 0);
	lstatus: std_logic_vector(15 downto 0);
	lcommand: std_logic_vector(15 downto 0);
end record;

type cfg_o_type is record
	dwaddr: std_logic_vector(9 downto 0); -- management dword address, adresses dword configuration register
	rd_en: std_logic; -- management read enable
	err_ur: std_logic; -- config error unsupported request
	err_cor: std_logic; -- config error correctable error
	err_ecrc: std_logic; -- ecrc error report
	err_cpl_timeout: std_logic; -- config error completion timeout
	err_cpl_abort: std_logic; -- config error completion aborted
	err_posted: std_logic; -- config error posted, errorneous transaction was a posted transaction
	err_locked: std_logic; -- config error locked, error was in locked transaction
	err_tlp_cpl_header: std_logic_vector(47 downto 0); -- error tlp completion header
	interrupt: std_logic; -- request signal interrupt
	interrupt_assert: std_logic; -- legacy interrupt assert/deassert select
	interrupt_di: std_logic_vector(7 downto 0); -- interrupt data in
	turnoff_ok: std_logic; -- application signals that turning off is OK
	pm_wake: std_logic; -- send PMPE Message, wake up upstream link partner
	trn_pending: std_logic; -- user transaction pending, must be asserted if a completion to a request is still pending
	dsn: std_logic_vector(63 downto 0); -- device serial number
end record;

end package types;
